import json
import logging
import os
import random
import sys
import traceback
import re
from datetime import datetime, timedelta, timezone

import boto3
import pyformance
import statsd
import wavefront_lambda
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

from expire import metrics

HOST_ES = os.environ.get('host', 'http://localhost:9200')
LQ_INDEX = os.environ.get('index', 'a_livequery')
LQ_DOC_TYPE = os.environ.get('doc_type', 'logs')
STATUS = os.environ.get('status', 'ACTIVE')
USER = os.environ.get('ES_USERNAME', None)
PASSWORD = os.environ.get('ES_PASSWORD', None)
BATCH_SIZE = int(os.environ.get('batch_size', 1000))
LOG_LEVEL = os.environ.get('log_level', 'info').upper()
NOTIFIER_FUNCTION = os.environ.get('NOTIFIER_FUNCTION', 'liveQueryStateNotifier')
ENV = os.environ.get('env', 'local')
REGION = os.environ.get('region', 'us-east-1')
LQ_EXPIRE_DEL_DAYS_BEFORE = int(os.environ.get('LQ_EXPIRE_DEL_DAYS_BEFORE', 15))  # delete the entry older than 15 days
SERVICE_NAME = 'livequery-expire'
HOSTNAME = os.environ.get('HOSTNAME', 'localhost-expire')
# Telegraf config
T_host = os.environ.get("TELEGRAF_HOST", "telegraf")
T_port = os.environ.get("TELEGRAF_PORT", 8125)
TELEGRAF_REPORTER_ENABLED = bool(os.environ.get('TELEGRAF_REPORTER_ENABLED', "False"))
TELEGRAF_REPORTER_TEST = bool(os.environ.get('TELEGRAF_REPORTER_TEST', "False"))
EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")


# Logging configuration
def handle_exception(exc_type, exc_value, exc_traceback):
    logger.error(traceback.format_tb(exc_traceback, 10))


logger = logging.getLogger(__name__)
handler = logging.StreamHandler(stream=sys.stdout)
file_formatter = logging.Formatter("{'@timestamp':'%(asctime)s', 'level': '%(levelname)s', 'message': '%(message)s'}")
handler.setFormatter(file_formatter)
logger.addHandler(handler)
logger.setLevel(logging.getLevelName(LOG_LEVEL))
sys.excepthook = handle_exception

# Create es client
logger.info('Creating ES client...')
if USER and PASSWORD:
    es = Elasticsearch(
        [HOST_ES],
        http_auth=(USER, PASSWORD),
        port=443,
        use_ssl=True,
        ssl_assert_hostname=False,
        request_timeout=30,
        retry_on_timeout=True
    )
else:
    es = Elasticsearch(HOST_ES, request_timeout=30, retry_on_timeout=True)

logger.info('Creating lambda function client...')
lambda_client = boto3.client('lambda')


def main():
    telegraf_client = None
    wavefront_registry = None
    current_time = int(datetime.utcnow().timestamp() * 1000)

    if TELEGRAF_REPORTER_ENABLED:
        telegraf_client = statsd.StatsClient(host=T_host, port=T_port, prefix=SERVICE_NAME)
    else:
        wavefront_registry = pyformance.MetricsRegistry()

    if TELEGRAF_REPORTER_TEST:
        metric_push(wavefront_registry, telegraf_client, f'{metrics.APIMetrics.ADMIN_NOTIFY}-test',
                    random.randint(1, 100))
        metric_push(wavefront_registry, telegraf_client, f'{metrics.APIMetrics.EXPIRED_COUNT}-test',
                    random.randint(1, 100))
        metric_push(wavefront_registry, telegraf_client, f'{metrics.APIMetrics.ADMIN_NOTIFY_FAIL}-test',
                    random.randint(1, 100))
        metric_push(wavefront_registry, telegraf_client, f'{metrics.APIMetrics.ADMIN_NOTIFY_PAYLOAD}-test',
                    random.randint(1, 100))

    logger.info('Searching for queries whose timeout times are less than {}'.format(current_time))
    logger.info('NOTIFIER_FUNCTION : {}'.format(NOTIFIER_FUNCTION))
    search_and_update_queries(current_time=current_time, wavefront_registry=wavefront_registry,
                              telegraf_client=telegraf_client)

    if LQ_INDEX == 'a_livequery' and LQ_DOC_TYPE == 'logs':
        # calculate the epoch time in GMT
        n_days_ago = datetime.now(timezone.utc) - timedelta(days=LQ_EXPIRE_DEL_DAYS_BEFORE)
        n_days_ago = int(n_days_ago.replace(hour=0, minute=0, second=0, microsecond=0).timestamp()) * 1000
        logger.info('older entries from calculated time {} to get delete'.format(n_days_ago))
        delete_non_lq_destination_execution_doc_from_livequery(day_before=n_days_ago)


def search_and_update_queries(current_time, wavefront_registry, telegraf_client):
    # Build query to find expired live queries
    query = build_es_query(current_time=current_time)
    logger.debug('Built ES query for expired live queries = {}'.format(query))

    # Search for queries
    page = es.search(index=LQ_INDEX, doc_type=LQ_DOC_TYPE, scroll='2m', search_type='query_then_fetch', size=BATCH_SIZE,
                     body=query)
    logger.debug('ES returned {} results'.format(page))

    sid = page['_scroll_id']
    scroll_size = page['hits']['total']

    # Start scrolling
    while scroll_size > 0:
        logger.info('Scrolling through result pages...')

        # Get number of results that were returned in the last scroll
        scroll_size = len(page['hits']['hits'])
        if scroll_size == 0:
            logger.info('Finished scrolling.')
            break

        logger.debug('Scroll size = {}'.format(scroll_size))

        queries = page['hits']['hits']

        # Collect query ids
        query_ids = [hit['_id'] for hit in queries]

        # Expire this set of query ids
        logger.debug('Expiring queries for ids = {}'.format(query_ids))
        set_expired(query_ids, wavefront_registry, telegraf_client)

        # Notify admins
        try:
            logger.info('Notifying live query admins.')
            logger.debug('Building notification payloads for live query requests {}'.format(queries))
            notify_payloads = []

            for query in queries:
                query_source = query["_source"]
                if query_source["notifyOnFinish"]:
                    try:
                        payload = build_notifier_payload(live_query_request=query_source)
                        if isEmail(payload["admin"]):
                            notify_payloads.append(payload)
                    except Exception as e:
                        metric_push(wavefront_registry, telegraf_client, metrics.APIMetrics.ADMIN_NOTIFY_FAIL)
                        logger.error(f"Failed to build notification payload for query source = {query_source}. "
                                     "Skipping notification for this query.")

            logger.info('Built notifier payloads = {}'.format(notify_payloads))

            for payload in notify_payloads:
                notify(payload=payload, wavefront_registry=wavefront_registry, telegraf_client=telegraf_client)
        except Exception as e:
            metric_push(wavefront_registry, telegraf_client, metrics.APIMetrics.ADMIN_NOTIFY_FAIL)
            logger.error("Errors occurred during email notifications. {}.", format(e))

        # Scroll and update scroll ID
        page = es.scroll(scroll_id=sid, scroll='2m')
        sid = page['_scroll_id']


def set_expired(ids, wavefront_registry, telegraf_client):
    """ Takes list of expired queries, updates their state to TIMED_OUT and re-indexes them. """
    update_time = int(datetime.utcnow().timestamp() * 1000)

    update_query_template = {
        "_op_type": "update",
        "_index": LQ_INDEX,
        "_type": LQ_DOC_TYPE,
        "doc": {
            "status": "TIMED_OUT",
            "statusUpdateTime": int(update_time)
        }
    }

    bulk_updates = []

    for id in ids:
        temp = update_query_template.copy()
        temp["_id"] = id
        bulk_updates.append(temp)

    logger.info('Bulk updating live queries.')
    logger.debug('Updates = {}'.format(bulk_updates))

    expired_time = int(datetime.utcnow().timestamp() * 1000)
    res = bulk(es, bulk_updates)
    logger.debug('Bulk update results = {}'.format(res))

    # send expired count to wavefront
    try:
        count = len(ids)
        metric_push(wavefront_registry, telegraf_client, 'LIVE_QUERY.expired', count)
        # wavefront_registry.counter('LIVE_QUERY.expired').inc(count)

        logger.info('Sent metrics to wavefront with count of {}'.format(count))
    except Exception as e:
        logger.error("Failed to send metrics to wavefront.")


def build_es_query(current_time):
    query = {
        "sort": [
            {
                "createTime": {
                    "order": "asc"
                }
            }
        ],
        "query": {
            "bool": {
                "must": {
                    "match": {
                        "status": STATUS
                    }
                },
                "filter": {
                    "range": {"timeoutTime": {"lte": current_time}}
                }
            }
        },
        "_source": ["_id", "createdBy", "createTime", "sql", "id", "notifyOnFinish", "name", "orgKey"]
    }

    return query


def build_es_delete_query(day_before):
    query = {
        "query": {
            "query_string": {
                "query": "_exists_:destinations AND (destinations:VULN OR destinations:COMPL) AND createTime:<{}".format(
                    day_before),
            }
        }
    }
    return query


def notify(payload, wavefront_registry, telegraf_client):
    """Invokes notifier function"""
    logger.debug("Invoking {} for event {}".format(NOTIFIER_FUNCTION, payload))
    invoke_response = lambda_client.invoke(FunctionName=NOTIFIER_FUNCTION,
                                           InvocationType='Event',
                                           Payload=json.dumps(payload))
    metric_push(wavefront_registry, telegraf_client, metrics.APIMetrics.ADMIN_NOTIFY)
    logger.debug("Notify response: {}".format(invoke_response))


def isEmail(email):
    if not EMAIL_REGEX.match(email):
        return False
    else:
        return True


def build_notifier_payload(live_query_request):
    """Builds notification payload from a live query request object"""
    return {
        "admin": live_query_request['createdBy'],
        "orgKey": live_query_request['orgKey'],
        "query": live_query_request['sql'],
        "queryId": live_query_request['id'],
        "name": live_query_request['name'],
        "startTime": {
            "nano": 0,  # this doesn't affect the notifier in any way
            "epochSecond": live_query_request['createTime'] / 1000
        },
        "status": "TIMED_OUT"
    }


# function deletes the non LQ destination execution from livequery ES index older than 15 days
# Destination types: LQ, VULN, COMPL
def delete_non_lq_destination_execution_doc_from_livequery(day_before):
    try:
        # Build query to delete non-LQ execution records which is older than 15 days
        query = build_es_delete_query(day_before=day_before)
        logger.debug('Built ES query for deleting non-LQ livequery executions = {}'.format(query))

        # Search for queries, we don't require data instead we just need total doc count with given query
        result = es.search(index=LQ_INDEX, doc_type=LQ_DOC_TYPE, body=query, track_total_hits=True, size=0)
        logger.info('Total number of non-LQ destination records going to delete are [{}] before'
                    ' time: [{}]'.format(result['hits']['total'], day_before))

        if result['hits']['total'] != 0:
            result = es.delete_by_query(index=LQ_INDEX, doc_type=LQ_DOC_TYPE, body=query, wait_for_completion=False)
            logger.debug('ES returned [{}] results'.format(result))

    except Exception as e:
        logger.error("Errors occurred during non LQ destination execution, exception {}.", format(e))
        handle_exception(exc_type=type(e), exc_value=e, exc_traceback=traceback.format_exc())


def metric_push(wavefront_registry, telegraf_client, counter_name, count=1):
    if TELEGRAF_REPORTER_ENABLED:
        telegraf_counter(telegraf_client, counter_name, count)
    else:
        wavefront_counter(wavefront_registry, counter_name, count)


def wavefront_counter(wavefront_registry, counter_name, count):
    try:
        wavefront_registry.counter(counter_name).inc(count)
        logger.info('Sent wavefront metrics for %s count of 1', counter_name)
    except Exception as we:
        logger.exception("Failed to send metrics to wavefront. Error: {}".format(we))


def telegraf_counter(telegraf_client, counter_name, count=1, rate=1):
    try:
        telegraf_client.incr(counter_name, count, rate,
                             tags={"env": ENV, "service": SERVICE_NAME, "hostname": HOSTNAME, "region": REGION})
        logger.info('Sent telegraf client metrics for %s count of 1', counter_name)
    except Exception as we:
        logger.exception("Failed to send metrics to telegraf_client. Error : {}".format(we))


@wavefront_lambda.wrapper
def lambda_handler(event, context):
    main(wavefront_registry=wavefront_lambda.get_registry())
